/* -*- mode: C++; tab-width: 8; c-basic-offset: 8; c-file-offsets: ((substatement-open . 0)) -*- */

#include "Document.hpp"
#include "XercesDocumentBody.hpp"
#include "Location.hpp"
#include "log.hpp"

#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMText.hpp>
#include <xercesc/dom/DOMAttr.hpp>

using namespace XERCES_CPP_NAMESPACE;
using namespace std;

GETLOGGER("xmltio.Document");

namespace xmltio {
TIODocument::TIODocument(const std::string& xml)
	: boost::shared_ptr<XercesDocumentBody>(new XercesDocumentBody(xml))
{
}
TIODocument::TIODocument(const Location& l)
	: boost::shared_ptr<XercesDocumentBody>(new XercesDocumentBody(*l.doc_.get(), l.node_))
{
}
TIODocument::TIODocument(XercesDocumentBody* body)
	: boost::shared_ptr<XercesDocumentBody>(body)
{
}
TIODocument::TIODocument(XERCES_CPP_NAMESPACE::DOMDocument* doc)
	: boost::shared_ptr<XercesDocumentBody>(new XercesDocumentBody(doc))
{
}

TIODocument TIODocument::clone() const
{
	return TIODocument(new XercesDocumentBody(*get()));
}

Location TIODocument::import(const Location& l)
{
	Location::Node* n = get()->doc_->importNode(l.node_, true);

	get()->getNamespaces().merge(l.doc_->getNamespaces());
	return Location(*this, n);
}

Location TIODocument::getRootLocation()
{
	return Location(*this, get()->doc_->getDocumentElement());
}
const Location TIODocument::getRootLocation() const
{
	return Location(*this, get()->doc_->getDocumentElement());
}
ID TIODocument::getID() const
{
	return getRootLocation()["dbxml:id"].getTextValue();
}

Location TIODocument::createTextLocation(const std::string& val)
{
	try {
		DOMText* tn = get()->doc_->createTextNode(X(val));
		return Location(*this, tn);
	} catch(const DOMException& ex) {
		throw XMLTIO_Exception("couldn't create text location",
		                       ex);
	}
}

namespace {
inline string make_msg(const string& uri, const string& prefix,
                       const string& qname)
{
	ostringstream msg;
	msg << "couldn't create node " << qname
	    << '(' << prefix << '=' << uri;
	return msg.str();
}
}

Location TIODocument::createElement(const string& uri,
                                    const string& prefix,
                                    const string& qname)
{
	DEBUG("createElement(" << "uri=" << uri
	      << ", prefix=" << prefix
	      << ", qname=" << qname << ')');
	try {
		// add namespace declaration when prefix in use
		if(qname.find(":") != string::npos) {
			get()->nsps_.put(prefix, uri);
		}
		// create element -- W/OUT namespace!
		// otherwise, xerces puts a declaration on the node
		DOMElement* e = get()->doc_->createElement(X(qname));

		if(!e) {
			throw ParsingException(make_msg(uri, prefix,
			                                qname));
		}

		return Location(*this, e);
	} catch(const DOMException& ex) {
		throw XMLTIO_Exception(make_msg(uri, prefix, qname),
		                       ex);
	}
}

Location TIODocument::createAttribute(const string& uri,
                                      const string& prefix,
                                      const string& qname)
{
	try {
		DOMAttr* attr = NULL;
		if(qname.find(":") != string::npos) {
			get()->nsps_.put(prefix, uri);
			attr = get()->doc_->createAttributeNS(X(uri), X(qname));
		} else {
			attr = get()->doc_->createAttribute(X(qname));
		}

		if(!attr) {
			throw ParsingException(make_msg(uri, prefix,
			                                qname));
		}

		return Location(*this, attr);
	} catch(const DOMException& ex) {
		throw XMLTIO_Exception(make_msg(uri, prefix, qname),
		                       ex);
	}
}

/*	TIODocument::operator DOMDocument*()
	{
	std::cerr << "warning: operator DOMDocument* deprecated" << std::endl;
	return get()->doc_;
	}
	TIODocument::operator const DOMDocument*() const
	{
	std::cerr << "warning: operator DOMDocument* deprecated" << std::endl;
	return get()->doc_;
	}
*/
ostream& operator<<(ostream& out, const TIODocument& doc)
{
	if(doc.get()) {
		out << *(doc.get());
	} else {
		out << "uninitialized TIODocument";
	}

	return out;
}

}
