/* -*- mode: C++; tab-width: 8; c-basic-offset: 8; c-file-offsets: ((substatement-open . 0)) -*- */

#include "Location.hpp"
#include "XPath.hpp"
#include "log.hpp"
#include "exception.hpp"

#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMAttr.hpp>

#include <xqilla/xerces/XercesConfiguration.hpp>
#include <xqilla/utils/XPath2Utils.hpp>

using namespace std;
using namespace XERCES_CPP_NAMESPACE;

GETLOGGER("xmltio.location");

#define RETHROW_DOMEX(ex, msg) DEBUG("got DOM exception, code=" << ex.code); \
	switch(ex.code) { \
	case DOMException::NOT_SUPPORTED_ERR: \
	case DOMException::NO_MODIFICATION_ALLOWED_ERR: \
		throw xmltio::UnsupportedOperationException(msg); \
	default: \
		throw xmltio::XMLTIO_Exception(msg); \
	}


namespace xmltio {
const char* typestring(int type)
{
	switch(type) {
	case DOMNode::ELEMENT_NODE:
		return "element";
	case DOMNode::TEXT_NODE:
		return "text";
	case DOMNode::ATTRIBUTE_NODE:
		return "attribute";
	default:
		return "unknown";
	}
}
namespace {
string get_text_children(const Location& l)
{
	ostringstream oss;
	for(XPathIterator i = XPath("text()").evaluate(l);
	    i; ++i) {
		oss << NX(i->getNodeValue());
	}
	return oss.str();
}
}

DOMNode* Location::get_xpath(const Location& ctx, const XPath& xpath)
{
	DOMNode *n = const_cast<DOMNode*>(ctx.node_);
#if 0
	DEBUG("get_xpath(" << ctx << "[name="
	      << ctx.getName()
	      << "], " << xpath.getXPath()
	      << "); doc=" << ctx.doc_);
#endif
	XPathIterator r(xpath.evaluate(ctx.doc_, n));

	if(!r) {
		throw NotFoundException("no node found at ", xpath);
	}

	return (DOMNode*) r;
}

::Node::Ptr Location::createXQNode(DOMNode* node,
                                   DynamicContext* context) {
	static ::XercesConfiguration config;
	return config.createNode(node, context);
}

Location::Location(const TIODocument& doc, Node* n)
	: doc_(doc), node_(n)
{
	if(NULL == doc_.get()) {
		throw XMLTIO_Exception("document NULL");
	}
	if(NULL == node_) {
		node_ = doc_.getRootLocation().node_;
	}
}

Location::Location(const std::string& xml, const XPath& xpath)
	: doc_(parse(xml)),
	  node_(get_xpath(doc_.getRootLocation()[XPath("/")], xpath))
{
}

Location::Location(const TIODocument& d, const XPath& xpath)
	: doc_(d),
	  node_(get_xpath(doc_.getRootLocation()[XPath("/")], xpath))
{
}

Location::Location(const std::string& xml)
	: doc_(parse(xml)),
	  node_(get_xpath(doc_.getRootLocation(), XPath("/")))
{
}

Location::Location(const Location& ctx, const XPath& xpath)
	: doc_(ctx.doc_), node_(0)
{
	node_ = get_xpath(ctx, xpath);
}

DOMNode* Location::attr(const char* name) const
{
	DOMNamedNodeMap *attributes = node_->getAttributes();
	if(!attributes) {
		ostringstream oss;
		oss << "location " << NX(node_->getNodeName())
		    << " can't have attributes (looking for "
		    << name <<  ")";
		throw UnsupportedOperationException(oss.str());
	}

	return attributes->getNamedItem(X(name));
}

Location Location::operator[](const char* name)
{
	DOMNode *item = attr(name);

	// if no such attribute exists, create it
	if(NULL == item) {
		DEBUG("adding new attribute node '" << name << "'");
		if (node_->getNodeType() != DOMNode::ELEMENT_NODE)
			throw NotFoundException("parent node is not an element");
		try {
			item = node_->getOwnerDocument()->createAttribute(X(name));
			node_->getAttributes()->setNamedItem(item);
		} catch(const DOMException& ex) {
			switch(ex.code) {
			case DOMException::INVALID_CHARACTER_ERR:
				throw InvalidCharacterException(string("attribute name '") + name + "' invalid", ex);
			case DOMException::NO_MODIFICATION_ALLOWED_ERR:
				throw NotFoundException(ex);
			default:
				throw XMLTIO_Exception(NX(ex.msg));
			}
		}
	}

	return Location(doc_, item);
}
Location Location::operator[] (const char* name) const
{
	DOMNode *item = attr(name);
	if(!item)
		throw NotFoundException(string("attribute '") + name + "' not found in constant document");
	return Location(doc_, item);
}
Location Location::operator[](const XPath& xpath) const
{
	XPathIterator i = xpath.evaluate(*this);
	if(!i) {
		throw NotFoundException("no node at ", xpath);
	}
	return *i;
}

void removeChildren(DOMNode* n)
{
	// remove any children
	if(n->hasChildNodes())
	{
		DOMNode *child = n->getFirstChild();
		while(child) {
			DEBUG("removing child");
			n->removeChild(child);
			child->release();
			child = n->getFirstChild();
		}
	}
}

Location& Location::operator=(const std::string& val)
{
	try {
		switch(node_->getNodeType())
		{
		case DOMNode::TEXT_NODE:
		case DOMNode::ATTRIBUTE_NODE:
		case DOMNode::CDATA_SECTION_NODE:
			node_->setNodeValue(X(val));
			break;
		case DOMNode::DOCUMENT_NODE:
			throw UnsupportedOperationException("cannot assign text content to document node");
			break;
		default:
			node_->setTextContent(X(val));
			break;
		}
		return *this;
	} catch(const DOMException& ex) {
		RETHROW_DOMEX(ex, "not assigned");
	}
}

string Location::getTextValue() const
{
	try {
		switch(node_->getNodeType()) {
		case DOMNode::TEXT_NODE:
		case DOMNode::ATTRIBUTE_NODE:
		case DOMNode::CDATA_SECTION_NODE:
			return NX(node_->getNodeValue());
		case DOMNode::ELEMENT_NODE:
			return get_text_children(*this);
		case DOMNode::DOCUMENT_NODE:
			throw UnsupportedOperationException("cannot extract from document node");
			break;
		default:
			return NX(node_->getTextContent());
		}
	} catch(const DOMException& ex) {
		RETHROW_DOMEX(ex, "no value");
	}
}

Location::Node* Location::import(const Location& l)
{
	// prepare node for transfer: foreign nodes
	// are imported and nodes already in the tree
	// are cloned
	if(l.doc_ != doc_) {
		return doc_.import(l).node_;
	} else {
		if(l.node_->getParentNode()) {
			return l.node_->cloneNode(true);
		} else {
			return l.node_;
		}
	}
}

Location Location::createElement(const char* name) {
	if (node_->getNodeType() != DOMNode::ELEMENT_NODE)
		throw NotFoundException("parent node is not an element");
	DOMNode *item;
	try {
		item = node_->getOwnerDocument()->createElement(X(name));
		node_->appendChild(item);
	} catch(const DOMException& ex) {
		switch(ex.code) {
		case DOMException::INVALID_CHARACTER_ERR:
			throw InvalidCharacterException(string("element name '") + name + "' invalid", ex);
		case DOMException::NO_MODIFICATION_ALLOWED_ERR:
			throw NotFoundException(ex);
		default:
			throw XMLTIO_Exception(NX(ex.msg));
		}
	}
	return Location(doc_, item);
}

Location Location::add(const Location& l)
{
	try {
		Node* n = import(l);

		DEBUG("nodeType=" << n->getNodeType());
		switch(n->getNodeType()) {
		case DOMNode::ATTRIBUTE_NODE: {
			DOMNamedNodeMap* attrs =
				node_->getAttributes();
			if(!attrs) {
				ostringstream msg;
				msg << "node '" << l
				    << "' is attribute"
				    << " but parent '" << *this
				    << "' cannot support attributes";
				throw XMLTIO_Exception(msg.str());
			}
			DOMNode* previous = attrs->setNamedItem(n);
			if(previous && previous != n) {
				previous->release();
			}
			break;
		}
		default:
			n = node_->appendChild(n);
			break;
		}

		return Location(doc_, n);
	} catch(const DOMException& ex) {
		throw XMLTIO_Exception("add: ", ex);
	}
}


Location Location::addBefore(const Location& l, const Location& lref)
{
	try {
		Node* n = import(l);

		DEBUG("nodeType=" << n->getNodeType());
		switch(n->getNodeType()) {
		case DOMNode::ATTRIBUTE_NODE: {
			DOMNamedNodeMap* attrs =
				node_->getAttributes();
			if(!attrs) {
				ostringstream msg;
				msg << "node '" << l
				    << "' is attribute"
				    << " but parent '" << *this
				    << "' cannot support attributes";
				throw XMLTIO_Exception(msg.str());
			}
			DOMNode* previous = attrs->setNamedItem(n);
			if(previous && previous != n) {
				previous->release();
			}
			break;
		}
		default:
			n = node_->insertBefore(n,lref.node_);
			break;
		}

		return Location(doc_, n);
	} catch(const DOMException& ex) {
		throw XMLTIO_Exception("add: ", ex);
	}
}


void Location::remove(Location& l)
{
	try {
		switch(l.node_->getNodeType()) {
		case DOMNode::ATTRIBUTE_NODE: {
			DOMNamedNodeMap* attrs = node_->getAttributes();
			if(!attrs) {
				throw XMLTIO_Exception("trying to remove an attribute, but getAttributes() returns NULL");
			}
			attrs->removeNamedItem(l.node_->getNodeName());
			break;
		}
		default:
			node_->removeChild(l.node_);
			break;
		}
		//			l.node_->release();
	} catch(const DOMException& ex) {
		throw XMLTIO_Exception(ex);
	}
}

Location& Location::replace (const Location& l)
{
	try {
		Location p(this->getParent());
		p.remove(*this);
		*this = p.add(l);
		return *this;
	} catch(const DOMException& e) {
		throw XMLTIO_Exception(NX(e.msg));
	}
}

/*
  Location Location::replace(Location& old, const Location& n)
  {
  switch(old.node_->getNodeType()) {
  case DOMNode::ATTRIBUTE_NODE:
  add(n);	// for an attribute, its equivalent
  break;
  default:
  try {
  node_->replaceChild(old.node_, import(n));
  break;
  } catch(const DOMException& ex) {
  throw XMLTIO_Exception("replace: ", ex);
  }
  }
  }
*/

LocationPtr find(const Location& l, const XPath& xpath)
{
	XPathIterator i = xpath.evaluate(l);
	if(!i) {
		return LocationPtr();
	}

	return LocationPtr(new Location(l.getDocument(), i));
}

void Location::setID(const ID& id)
{
	(*this)["xml:id"] = id;
}

}
