/* -*- mode: C++; tab-width: 8; c-basic-offset: 8; c-file-offsets: ((substatement-open . 0)) -*- */
/**
 *
 * @file	TypeConversion.hpp
 * @author	Ingo Luetkebohle (iluetkeb@TechFak.Uni-Bielefeld.de)
 * @date	started at 2004/06/18
 * @version	$Id: TypeConversion.hpp 4275 2016-09-06 13:10:10Z rhaschke $
 *
 **/

#ifndef __XMLTIO_TYPECONVERSION_HPP
#define __XMLTIO_TYPECONVERSION_HPP

#include "str.hpp"
#include "exception.hpp"
#include "Location.hpp"
#include "Template.hpp"

#include <string>
#include <sstream>
#include <iomanip>
#include <fstream>

#include <stdlib.h>		// for strtol

#ifdef WIN32
#define strtoll _strtoi64
#define strtoull _strtoui64
#define strtof (float) strtod
#endif

namespace xmltio {

/** \brief convert T to Location (text node value).
 *
 * Usage is simple:
 * <pre>
 * Location l;
 * float f;
 * l = f;
 * </pre>
 *
 * Specializations for <code>string</code>, <code>long</code>
 * and <code>double</code> exist.  For user-defined types, a
 * template specialization must be provided.  Note that it has
 * to be placed into namespace <code>xmltio</code>!  <p> For
 * consistency reasons, if exceptions are thrown, they should
 * be derived from <code>XMLTIO_Exception</code>.
 *
 * @see extract
 */
// to-string converter specializations

// NOTE: it is tempting to provide a default implementation
// based on streams.  however, that prevents the compiler from
// using specialiations provided in other headers.

template<typename T>
std::string type_to_string(const T& value)
{
	std::ostringstream oss;
	oss.imbue(std::locale("C"));
	// should correspond approx. to maximum precision of doubles
	oss << std::setprecision(16) << value;
	return oss.str();
}

inline void convert(const long& value, Location& l)
{
	l = type_to_string(value);
}
inline void convert(const unsigned long& value, Location& l)
{
	l = type_to_string(value);
}
inline void convert(const long long& value, Location& l)
{
	l = type_to_string(value);
}
inline void convert(const unsigned long long& value, Location& l)
{
	l = type_to_string(value);
}
inline void convert(const int& value, Location& l)
{
	l = type_to_string(value);
}
inline void convert(const unsigned int& value, Location& l)
{
	l = type_to_string(value);
}
inline void convert(const double& value, Location& l)
{
	l = type_to_string(value);
}
inline void convert(const float& value, Location& l)
{
	l = type_to_string(value);
}
inline void convert(const std::string& x, Location& l)
{
	l = x;
}
inline void convert(const char* x, Location& l)
{
	l = std::string(x);
}

/** \brief convert Location to user type.
 *
 * Usage:
 * <pre>
 * Location l;
 * string value = extract<string>(l);
 * </pre>
 *
 * Specializations for string <code>string</code>,
 * <code>long</code> and <code>double</code> exist.  For
 * user-defined types, the "get" member function must be
 * specialized.
 *
 * @see convert
 **/
template<typename T>
struct extract {
	const Location loc_;
	extract(const Location& l) : loc_(l) {}
	T get() const;
	operator T() const {
		return get();
	}
	static T from_xml(const std::string& xml) {
		return extract(find_anchor<T>(xml)).get();
	}
};


// from-string converter specializations
template<> inline std::string extract<std::string>::get() const
{
	return loc_.getTextValue();
}
template<> inline long  extract<long>::get() const
{
	const std::string &value = loc_.getTextValue();
	const char* nptr = value.c_str();
	char* endptr;
	long l = strtol(nptr, &endptr, 10);
	if(nptr == endptr) {
		std::ostringstream buf;
		buf << "no number value found in '" << value << "'";
		throw ConversionFailedException(buf.str());
	}
	return l;
}
template<> inline unsigned long extract<unsigned long>::get() const
{
	const std::string &value = loc_.getTextValue();
	const char* nptr = value.c_str();
	char* endptr;
	unsigned long l = strtoul(nptr, &endptr, 10);
	if(nptr == endptr) {
		std::ostringstream buf;
		buf << "no number value found in '" << value << "'";
		throw ConversionFailedException(buf.str());
	}
	return l;
}
template<> inline long long extract<long long>::get() const
{
	const std::string &value = loc_.getTextValue();
	const char* nptr = value.c_str();
	char* endptr;
	long long l = strtoll(nptr, &endptr, 10);
	if(nptr == endptr) {
		std::ostringstream buf;
		buf << "no number value found in '" << value << "'";
		throw ConversionFailedException(buf.str());
	}
	return l;
}
template<> inline unsigned long long extract<unsigned long long>::get() const
{
	const std::string &value = loc_.getTextValue();
	const char* nptr = value.c_str();
	char* endptr;
	unsigned long long l = strtoull(nptr, &endptr, 10);
	if(nptr == endptr) {
		std::ostringstream buf;
		buf << "no number value found in '" << value << "'";
		throw ConversionFailedException(buf.str());
	}
	return l;
}
template<> inline int extract<int>::get() const
{
	return (int)extract<long>(loc_);
}
template<> inline unsigned int extract<unsigned int>::get() const
{
	return (unsigned int)extract<unsigned long>(loc_);
}
template<> inline double extract<double>::get() const
{
	const std::string &value = loc_.getTextValue();
	std::istringstream istr(value);
	istr.imbue(std::locale("C"));
	double v = 0.0;
	istr >> v;
	if((istr.rdstate() & std::ifstream::failbit) != 0 ) {
		std::ostringstream buf;
		buf << "no number value found in '" << value << "'";
		throw ConversionFailedException(buf.str());
	}
	return v;
}
template<> inline float extract<float>::get() const
{
	const std::string &value = loc_.getTextValue();
	std::istringstream istr(value);
	istr.imbue(std::locale("C"));
	float v = 0.0f;
	istr >> v;
	if((istr.rdstate() & std::ifstream::failbit) != 0 ) {
		std::ostringstream buf;
		buf << "no number value found in '" << value << "'";
		throw ConversionFailedException(buf.str());
	}
	return v;
}
}

#endif //  __XMLTIO_TYPECONVERSION_HPP
