/* -*- mode: C++; tab-width: 8; c-basic-offset: 8; c-file-offsets: ((substatement-open . 0)) -*- */
/**
 *
 * @file	XPath.hpp
 * @author	Ingo Luetkebohle (iluetkeb@TechFak.Uni-Bielefeld.de)
 * @date	started at 2004/06/18
 * @version	$Id: XPath.hpp 4251 2015-01-14 21:35:25Z rhaschke $
 *
 **/

#ifndef __XMLTIO_XPATH_HPP
#define __XMLTIO_XPATH_HPP

#include "Location.hpp"
#include "str.hpp"
#include "xmltio_config.hpp"

#include <memory>
#include <string>
#include <iosfwd>
#include <xercesc/dom/DOMNode.hpp>
#include <xqilla/context/DynamicContext.hpp>
#include <xqilla/simple-api/XQQuery.hpp>
#include <xqilla/runtime/Result.hpp>
#include <boost/shared_ptr.hpp>
#include <pthread.h>

namespace xmltio {

typedef boost::shared_ptr<DynamicContext> DynamicContextPtr;
typedef boost::shared_ptr<XQQuery> XQQueryPtr;

class XMLTIO_API XPathIterator {
	friend class XPath;
private:
	typedef XERCES_CPP_NAMESPACE::DOMNode Node;
	/// a reference to the originating document, so that
	/// it will be kept around as long as the iterator
	/// instance exists
	TIODocument       doc_;
	DynamicContextPtr ctx_;
	mutable Result    res_;
	Item::Ptr         itm_;

	void  next();

	XPathIterator(const TIODocument& doc, const Result& r,
	              const DynamicContextPtr& c);
protected:
	Node* item() const;

public:
	XPathIterator();

	XPathIterator(const XPathIterator& o);

	XPathIterator& operator++() {
		next();
		return *this;
	};

	~XPathIterator();

	Location operator*() {
		return Location(doc_, item());
	}

	operator Node*() {
		return item();
	}

	Node* operator->() {
		return item();
	}

	XPathIterator& operator=(const XPathIterator& o);

	bool operator!=(const XPathIterator& o) const
	{
		return item() != o.item();
	}

	bool operator!() const {
		return (NULL == item());
	}
	operator bool() {
		return (NULL != item());
	}

};

/** \brief helper class to simplify XPath evaluation.
 */
class XMLTIO_API XPath {
private:
	std::string xpath_;
	mutable pthread_mutex_t   contextMutex_;
	DynamicContextPtr context_;
	mutable XQQueryPtr query_;

	void  compile();

public:
	/** \brief Create XPath evaluator.
	 *
	 * @param xpath XPath expression
	 * @throw MalformedXPathException when expression is empty
	 */
	XPath(const char* xpath);
	XPath(const std::string& xpath);
	XPath(const XPath& xpath);
	~XPath();

	/** \brief run match against given node.
	 * @return true when at least one result is found
	 */
	bool isMatch(const Location& l) const;

	/** \brief evaluate XPath expression on given node.
	 * @param location the context node
	 */
	XPathIterator evaluate(const Location& location) const;
	XPathIterator evaluate(const TIODocument& d, XERCES_CPP_NAMESPACE::DOMNode*) const;

	bool operator==(const XPath& o) const {
		return xpath_ == o.xpath_;
	}
	bool operator!() const {
		return "" == xpath_;
	}

	XPath& operator=(const XPath& o) {
		xpath_ = o.xpath_;
		compile();
		return *this;
	}
	XPath& operator=(const std::string& xpath) {
		xpath_ = xpath;
		compile();
		return *this;
	}

	const std::string& getXPath() const { return xpath_; }

	friend std::ostream& operator<<(std::ostream&, const XPath&);
	friend std::istream& operator>>(std::istream&, XPath&);
	friend void convert(const XPath&, Location&);
};
typedef boost::shared_ptr<XPath> XPathPtr;

std::ostream& operator<<(std::ostream&, const XPath&);
std::istream& operator>>(std::istream&, XPath&);
}

#endif //  __XMLTIO_XPATH_HPP
